//
//  CEExtension.m
//  CESnippet
//
//  Created by 舘 康一 on 11/09/04.
//  Copyright 2011 Odoru-saboten Workshop. All rights reserved.
//

#import "CESController.h"
#import "CESnippet.h"
#import <objc/objc-class.h>

void Swizzle(Class c, SEL orig, SEL new)
{
    // オリジナルのメソッドコールを新しいメソッドに置き換える、新しいメソッドコールはオリジナルのものに置き換わる
    
    Method origMethod = class_getInstanceMethod(c, orig);
    Method newMethod = class_getInstanceMethod(c, new);
    
    class_addMethod(c, 
    				orig, 
                    method_getImplementation(newMethod), 
                    method_getTypeEncoding(newMethod));
    
    class_addMethod(c, 
    				new, 
                    method_getImplementation(origMethod), 
                    method_getTypeEncoding(origMethod));
    
    method_exchangeImplementations(origMethod, newMethod);
}

@implementation CESController

+ (void)load
{
	LOG(@"CESController loaded!");

	// ---------------------------------------------------------------------------------------------
	// メソッド Swizzing
	// ---------------------------------------------------------------------------------------------
    Class class = objc_getClass(CES_TEXT_VIEW_CLASS_NAME);
    
    // 初期化処理の置き換え
    Swizzle(class, @selector(initWithFrame:textContainer:), @selector(CESInitWithFrame:textContainer:));
    
    // タブ入力イベントの置き換え
    Swizzle(class, @selector(insertTab:), @selector(CESInsertTab:));
    
    // 終了処理の置き換え
    Swizzle(class, @selector(dealloc), @selector(CESDealloc));
    
	// ---------------------------------------------------------------------------------------------
    // メソッド追加
	// ---------------------------------------------------------------------------------------------
//    Class editorViewClass = objc_getClass(CES_EDITOR_VIEW_CLASS_NAME);
    
    // 
//    Method lineTextMethod = class_getInstanceMethod(editorViewClass, @selector(lineText:));
//    class_addMethod(editorViewClass, 
//                    lineTextMethod, 
//                    method_getImplementation(lineTextMethod),
//                    method_getTypeEncoding(lineTextMethod));
    
	// ---------------------------------------------------------------------------------------------
    // その他、このプラグインの初期化処理を行う
	// ---------------------------------------------------------------------------------------------
    
	// ---------------------------------------------------------------------------------------------
    // スニペットフォルダが存在するかどうかを確認し、存在しなければリソース内のデフォルトスニペットフォルダを
    // コピーする。
	// ---------------------------------------------------------------------------------------------
    NSFileManager *theFileManager = [NSFileManager defaultManager];
    
    NSString *snipPath = [NSHomeDirectory( ) 
                          stringByAppendingPathComponent:CES_SNIPPETS_FOLDER];
                          
    // アプリケーションサポートフォルダに、既存の Snippets フォルダがあるかどうかを確認する
    BOOL theBoolIsDir = NO;
    if (![theFileManager fileExistsAtPath:snipPath isDirectory:&theBoolIsDir]) {
    
    	LOG(@"Start Copy Snippets folder.");
        
        // 存在しなければ、自身のリソースの中からフォルダをコピーする
        NSString *bundlePath = [[NSBundle bundleForClass:[CESController class]] resourcePath];
        NSString *defaultSnipFolderPath = [NSString stringWithFormat:@"%@%@", bundlePath, @"/Snippets/"];
        
        BOOL copyResult = NO;
        copyResult = [theFileManager copyPath:defaultSnipFolderPath toPath:snipPath handler:nil];
        
        if (!copyResult) {
            NSLog(@"フォルダのコピーに失敗しました。");
        }
    }
    
}

@end




