//
//  CESTextCoreExtension.m
//  CESnippet
//
//  Created by 舘 康一 on 11/09/04.
//  Copyright 2011 Odoru-saboten Workshop. All rights reserved.
//

#import "CESTextCoreExtension.h"
#import "CESController.h"
#import "CESnippet.h"

@implementation NSTextView(CESTextCoreExtension)

#pragma mark === Swizzling Method ===

// ------------------------------------------------------------------
// CETextViewCore の初期化処理にスニペットの初期化処理を割り込ませる。
- (id)CESInitWithFrame:(NSRect)inFrame textContainer:(NSTextContainer *)inTextContainer
// ------------------------------------------------------------------
{
    LOG(@"CESInitWithFrame:textContainer:");
    
    // CotEditor オリジナルの初期化処理を実行
    self = [self CESInitWithFrame:inFrame textContainer:inTextContainer];
    
    // スニペットの初期化処理を実行
    [self CESSnippetObject];
    
    return self;
}

// ------------------------------------------------------------------
// CETextViewCore のタブ挿入イベントに処理を追加
- (void)CESInsertTab:(id)sender
// ------------------------------------------------------------------
{
	LOG(@"--- CESInsertTab ---");
    BOOL isExecuted = CES_SNIPPET_NOT_EXECUTED;
    
    @try {
        // Snippet を実行する
        isExecuted = [[self CESSnippetObject] executeTabTrigger:sender];
    }
    @catch (NSException * e) {
        NSLog(@"*CESInsertTab: Caught %@: %@", [e name], [e reason]);
    }
    
    // Snippet の実行が無ければ CotEditor オリジナルの insertTab をコールする
    if (!isExecuted) {
    	LOG(@"Snippet is not expand.");
        [self CESInsertTab:sender];
    }
}

// ------------------------------------------------------------------
// CETextViewCore の解放処理にスニペット解放処理を割り込ませる。
- (void)CESDealloc
// ------------------------------------------------------------------
{
    // スニペットの解放処理
    [self _removeSimulatedIVars];
    
    // CotEditor オリジナルの dealloc をコールする
    [self CESDealloc];
}


#pragma mark === 疑似インスタンス変数の設定 ===
//http://blog.boreal-kiss.com/2010/03/07000052.html/

// ------------------------------------------------------------------
// 擬似的にインスタンス変数に Snippet のオブジェクトを格納する
// ------------------------------------------------------------------

// 疑似インスタンス変数格納用辞書と、そのキーのための文字列
static NSMutableDictionary *_simulatedIVars = nil;
static NSString * const CESSnippetObjectForKeyString = @"";

// ------------------------------------------------------------------
// 疑似インスタンス変数を作るクラスメソッドをカテゴリに追加
+(NSMutableDictionary *)_simulatedIVars
// ------------------------------------------------------------------
{
	if (_simulatedIVars == nil){
		_simulatedIVars = [[NSMutableDictionary alloc] init];
	}
	return _simulatedIVars;
}

// ------------------------------------------------------------------
// インスタンスそのものの文字列表現を返す
-(NSString *)_instanceID
// ------------------------------------------------------------------
{
	return [NSString stringWithFormat:@"%@", self];
}

// ------------------------------------------------------------------
// 疑似インスタンス変数へのキーとなる文字列を生成して返す
-(NSString *)_objectKeyForKeyString:(NSString *)aKeyString
// ------------------------------------------------------------------
{
	return [NSString stringWithFormat:@"%@:%@", 
            [self _instanceID], aKeyString];
}

// ------------------------------------------------------------------
// インスタンス解放時に、疑似インスタンス変数に格納されたオブジェクトを回避する。
-(void)_removeSimulatedIVars
// ------------------------------------------------------------------
{
	NSString *aKey = 
    [self _objectKeyForKeyString:CESSnippetObjectForKeyString];
	id object = [[[self class] _simulatedIVars] objectForKey:aKey];
    
	if (object){
		[[[self class] _simulatedIVars] removeObjectForKey:aKey];
	}
}

// ------------------------------------------------------------------
// 疑似インスタンス変数に格納されたスニペットオブジェクトを返す
-(CESnippet*)_CESSnippetObjectForKeyString
// ------------------------------------------------------------------
{
	NSString *aKey = 
    [self _objectKeyForKeyString:CESSnippetObjectForKeyString];
	id object = [[[self class] _simulatedIVars] objectForKey:aKey];
    
	return object;
}

// ------------------------------------------------------------------
// スニペットオブジェクトを疑似インスタンス変数に格納する
-(void)_setCESSnippetObjectForKeyString:(CESnippet*)_snippet
// ------------------------------------------------------------------
{
	NSString *aKey = 
    [self _objectKeyForKeyString:CESSnippetObjectForKeyString];
	[[[self class] _simulatedIVars] 
     setObject:_snippet forKey:aKey];
}

// ------------------------------------------------------------------
- (CESnippet*)CESSnippetObject
// このテキストビューで使うスニペットオブジェクトを取得する。
// ------------------------------------------------------------------
{
    if (![self _CESSnippetObjectForKeyString]){
        // まだ初期化されていなければ初期化する。
		[self _setCESSnippetObjectForKeyString:[CESnippet initSnippet:self]];
	}
    
    return [self _CESSnippetObjectForKeyString];
}

@end
