//
//  CESnippet.m
//  CESnippet
//
//  Created by 舘 康一 on 11/09/05.
//  Copyright 2011 Odoru-saboten Workshop. All rights reserved.
//

#import "CESnippet.h"
#import <OgreKit/OgreKit.h>

@interface CESnippet (Private)
- (NSString *)pathOfSnippetsDirectory;
- (void)initInstance:(id)textView;
- (void)initSyntaxSnippet;
- (void)readSnippetFolder:(NSString*)theDirPath;
- (NSString *)findAbbreviation;
- (BOOL)expandAbbreviation:(NSString*)abbreviation;
- (NSString *)translateShellScript:(NSString*)snippet;
- (NSString *)executeShellScript:(NSString *)commandBaseString;
- (NSString *)translateVariables:(NSString*)snippet;
- (NSString *)replaceVariable:(NSString*)variableString;
- (NSString *)translateTabStop:(NSString*)snippet;
- (NSString *)varExecute:(NSString*)varName;
@end

@implementation CESnippet

#pragma mark Public Method

//------------------------------------------------------
- (BOOL)modeActivate
// 現在プレイスホルダ移動モードになっているかどうかを返す
//------------------------------------------------------
{
    return _isModeActivate;
}

//------------------------------------------------------
- (BOOL)executeTabTrigger:(id)sender
// タブトリガーを実行する
//------------------------------------------------------
{
    LOG(@"executeTabTrigger!");
    
    //-----------------------------------------------
    // Snippet モードになっていれば「次のプレイスホルダー」を実行する
    // Shiftキーが押されていたかどうかで挙動が違うが、その判定は呼び出し先のメソッドで行う。
    //-----------------------------------------------
    if (_isModeActivate) {
        // TODO: [self nextPlaceHolder:sender];
        
        // プレイスホルダー遷移処理を実行したため「実行した」を返し、タブ入力をキャンセルさせる
        return CES_SNIPPET_EXECUTED;
    }
	// Snippet モードでなければ以降の処理を行い、省略形の展開を試みる
    //-----------------------------------------------
    
    //-----------------------------------------------
    // スニペットを初期化する。シンタックスが設定されていないか変更されていた場合。
    //-----------------------------------------------
    [self initSyntaxSnippet];
    
    //-----------------------------------------------
    // 現在のキャレット位置から前方検索し、省略形を探す。
    //-----------------------------------------------
    NSString* abbreviation = [self findAbbreviation];
    [theTabTrigger setString:abbreviation];
    if ([abbreviation isEqualToString:@""]) {
    	// 省略形が見つからない場合「実行していない」を返す。これで通常のタブ入力となる。
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    //-----------------------------------------------
    // 省略形が、スニペットの初期化を促すものであれば初期化を行う
    //-----------------------------------------------
    if ([abbreviation isEqualToString:CES_UPDATE_SNIPPETS_COMMAND]) {
    	LOG(@"UPDATE_SNIPPETS_COMMAND");
        [self initSyntaxSnippet];
    }
    
    //-----------------------------------------------
    // 省略形を展開する
    //-----------------------------------------------
    BOOL successAbbreviation = [self expandAbbreviation:abbreviation];
    if (!successAbbreviation) {
    	// 省略形の展開が失敗した場合「実行していない」を返す。これで通常のタブ入力となる。
        LOG(@"Failed expand.");
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    //-----------------------------------------------
    // いくつかのプレイスホルダーが見つかっていればモードを Snippet モードにする
    //-----------------------------------------------
    // havePlaceHolder
    // [self 
    // _isModeActivate = YES;
    
    
    //-----------------------------------------------
    // 省略形が見つからない等、スニペットの実行が無ければ「実行していない」を返す。
    // 処理がここまで進んでいればスニペットの展開に問題がないと判断し、タブ入力イベントを
    // キャンセルさせるため「実行した」を返す。
    //-----------------------------------------------
    return CES_SNIPPET_EXECUTED;
}

//------------------------------------------------------
- (void)dealloc
// インスタンスの終了処理を行う
//------------------------------------------------------
{
    [snippetsDict release];
    [theTabTrigger release];
    [super dealloc];
}

#pragma mark Private Method

//------------------------------------------------------
// タブトリガーを展開する
- (BOOL)expandAbbreviation:(NSString*)abbreviation
//------------------------------------------------------
{
    NSString* snippet = [snippetsDict objectForKey:abbreviation];
    
    // 省略形（トリガー）が存在するかどうかを調べる
    if (!snippet) {
    	LOG(@"Tabtrigger Not Found in Array.");
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    // -------------------------------------------------------------------------
    // スニペット中のスクリプトや変数の展開を実行していく
    // -------------------------------------------------------------------------
    resultString = [NSMutableString stringWithString:snippet];
    
    // ------------------------------------------
    // 変数の置き換え
    // ------------------------------------------
    [resultString setString:[self translateVariables:resultString]];
    if (nil == resultString) {
    	LOG_TRACE;
        return CES_SNIPPET_NOT_EXECUTED;
    }
    // エスケープされた $ を置き換える
    [resultString setString:[resultString stringByReplacingOccurrencesOfString:@"\\$" withString:@"$"]];
    
    // ------------------------------------------
    // シェルの実行
    // ------------------------------------------
	[resultString setString:[self translateShellScript:resultString]];
    if (nil == resultString) {
    	LOG_TRACE;
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    // ------------------------------------------
    // プレースホルダの置き換え
    // 置き換え後に各プレースホルダの位置を保存する
    // ------------------------------------------
    
    // タブストップ（$0のみ）を置き換え
	[resultString setString:[self translateTabStop:resultString]];
    if (nil == resultString) {
    	LOG_TRACE;
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    // -------------------------------------------------------------------------
    // スニペットが完成したら省略形を置き換える
    // -------------------------------------------------------------------------
    [ceTextView setSelectedRange:abbreviationRange];
    [ceTextView insertText:resultString];
    
    // タブストップが指定されていればその位置を選択状態とする
    if ((int)0 <= (int)cursorPositionRange.location) {
        [ceTextView setSelectedRange:cursorPositionRange];
    }
    
    // ここまで来ていれば省略形の展開を終えているはずなので「実行した」ことを呼び出し元に知らせる
    return CES_SNIPPET_EXECUTED;
}

//------------------------------------------------------
// 現在のキャレット位置の左側に「省略形」を探す
- (NSString *)findAbbreviation
//------------------------------------------------------
{
    // 現在のキャレット位置を得る
    NSRange theSelectedRange = [ceTextView selectedRange];
    
    // 現在行の開始位置から現在のキャレットまでの文字列を得る
    NSRange rangeFromLineHeadToCaret = [[ceTextView string] lineRangeForRange:theSelectedRange];
    
    NSUInteger theLength = theSelectedRange.location - rangeFromLineHeadToCaret.location;
    NSString* theSelectedString = [[ceTextView string] substringWithRange:NSMakeRange(rangeFromLineHeadToCaret.location, theLength)];
    
    LOG(theSelectedString);
    
    // キャレット位置から前方に登録してあるタブトリガーを探す
    BOOL isFound = NO;
    NSUInteger tabTriggerLength;
    NSRange lastRange;
    NSMutableString* lastString = [NSMutableString string];
    for (id key in snippetsDict) {
        
        tabTriggerLength = [key length];
        lastRange = NSMakeRange([theSelectedString length] - tabTriggerLength, tabTriggerLength);
        
        if ([theSelectedString length] < lastRange.location + lastRange.length) {
            continue;
        }
        if ((int)lastRange.location < 0 || (int)lastRange.length <= 0) {
            continue;
        }
        
        [lastString setString:[theSelectedString substringWithRange:lastRange]];
        
        if ([lastString isEqualToString:key]) {
            
            // 最後の文字が英数字であれば単語の途中でありタブトリガーではないと判断する
            if (0 < lastRange.location) {
                NSRange beforeRange = NSMakeRange(lastRange.location - 1, 1);
                NSString *beforeChar = [NSString stringWithString:[theSelectedString substringWithRange:beforeRange]];
                
                OGRegularExpression* oniRegrex = [OGRegularExpression regularExpressionWithString:@"^[0-9a-zA-Z]"];
                OGRegularExpressionMatch* theMatche = [oniRegrex matchInString:beforeChar];
                if (nil != theMatche) {
                    continue;
                }
            }
            
        	// 見つかった
            isFound = YES;
            
            // 省略形が見つかった位置(NSRange)をインスタンス変数に保存する
            abbreviationRange = NSMakeRange(theSelectedRange.location - (int)tabTriggerLength, tabTriggerLength);
            
            break;
        }
    }
    
    // 省略形が無ければ空文字列を返す。これで通常のタブ入力となる。
    if (!isFound) {
        return @"";
    }
    
    // 得られた「省略形」を返す
    LOG(@"findAbbreviation end.");
    return lastString;
}

#pragma mark -- PlaceHolders --

//------------------------------------------------------
// スニペット中のタブストップ（$0）を置き換え、その位置を
// cursorPositionRange に保存する。
// ※ ${0:stop_pos} にはまだ対応していない。
- (NSString *)translateTabStop:(NSString*)snippet
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];
    
	// $0 に囲まれた文字列を抽出する正規表現
    NSString *regrex2 = @"(?<!\\\\)(\\$0){1}\\b";
    OGRegularExpression* oniRegrex2 =
    [OGRegularExpression regularExpressionWithString:regrex2];
    
    // Snippet を正規表現で検索し、最初のタブストップの位置を得る
    OGRegularExpressionMatch* theMatch = [oniRegrex2 matchInString:snippet];
    if (theMatch == nil) {
    	// マッチしなければ何もしない
        cursorPositionRange.location = -1;
        return snippet;
    }
    
    // $0 の位置を保存
    cursorPositionRange = [theMatch rangeOfMatchedString];
    // 置き換えるので長さはゼロに
    cursorPositionRange.length = 0;
    // 置き換え位置を、テキスト全体の位置に
    cursorPositionRange.location += abbreviationRange.location;
    
    // $0 を置き換え
    [retString appendString:[oniRegrex2 replaceFirstMatchInString:snippet
                                                       withString:@""]];
    return retString;
}

#pragma mark -- Shell Script --

//------------------------------------------------------
// スニペットの中のシェルスクリプトを探し出し、順に処理する。
// スクリプト部分を実行結果で置き換えるが、実行に失敗した場合
// 空文字列で置き換える。
- (NSString *)translateShellScript:(NSString*)snippet
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];

	// ` ` に囲まれた文字列を抽出する正規表現
    NSString *regrex2 = @"(?<!\\\\)`{1}.*?(?<!\\\\)`{1}";
    OGRegularExpression* oniRegrex2 =
    	[OGRegularExpression regularExpressionWithString:regrex2];
    
    // Snippet を正規表現で検索し、シェルスクリプト部分を得る
    NSEnumerator *theEnumerator;
    theEnumerator = [oniRegrex2 matchEnumeratorInString:snippet];
    NSArray *theMatchedAry = [oniRegrex2 allMatchesInString:snippet];
    
    if ([theMatchedAry count] == 0) {
    	LOG(@"Shell no execute.");
        return snippet;
    }
    
    // 全てのシェルを置き換える
    NSMutableString* lastPostmatchString = [NSMutableString stringWithString:@""];
    for (OGRegularExpressionMatch* theMatch in theMatchedAry) {
    
    	LOG(@"Start execute -> %@", [theMatch matchedString]);
    
        // マッチ部分とマッチ部分の間の文字列を接続する
        [retString appendString:[theMatch stringBetweenMatchAndLastMatch]];
    
        // マッチ部分をシェルスクリプトとして実行し、接続する
        [retString appendString:[self executeShellScript:[theMatch matchedString]]];
        
        // マッチ部分の後ろの文字列を保存する
        [lastPostmatchString setString:[theMatch postmatchString]];
    }
    // 最後のマッチ部分の後ろの文字列を接続する
    [retString appendString:lastPostmatchString];
    
    return retString;
}

//------------------------------------------------------
// シェルスクリプトを実行する
- (NSString *)executeShellScript:(NSString *)commandBaseString
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];
    
    // コマンド部分を取り出す（すなわち最初と最後の文字「`」を削除する）
    if ([commandBaseString length] < 3) {
        return @"";
    }
    NSRange commandRange = NSMakeRange(1, [commandBaseString length] - 2);
    NSString *commandString = [[NSString stringWithString:[commandBaseString substringWithRange:commandRange]]stringByReplacingOccurrencesOfString:@"\\`" withString:@"`"];
    
    // コマンド部分を実行する
    NSTask *task  = [[NSTask alloc] init];
    NSPipe *pipe  = [[NSPipe alloc] init];
    
    [task setLaunchPath: @"/bin/sh"];
    [task setArguments: [NSArray arrayWithObjects: @"-c", commandString, nil]];
    
    @try {
    	LOG(commandString);
        
        [task setStandardOutput: pipe];
        [task launch];
        
        LOG(@"task complete");
        
        NSFileHandle *handle = [pipe fileHandleForReading];
        NSData *data = [handle  readDataToEndOfFile];
        
        LOG(@"data complete");
        
        [retString setString:[[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease]];
        
        if (retString != nil) {
            if (![retString isEqualToString:@""]) {
                // 最後の文字が改行コードの場合、それを削除する
                NSRange lastRange = NSMakeRange([retString length] - 1, 1);
                
                NSString *lastChar = [NSString stringWithString:[retString substringWithRange:lastRange]];
                if ([lastChar isEqualToString:@"\n"]) {
                    [retString setString:
                     [retString substringWithRange:NSMakeRange(0, [retString length] -1)]];
                }
            }
		}
    }
    @catch (NSException * e) {
    	// 実行に失敗したら空文字列を返す。
        NSLog([e description]);
        [retString setString: @""];
    }
    @finally {
        [task release];
        [pipe release];
    }
    return retString;
}

#pragma mark -- init --

//------------------------------------------------------
+ (CESnippet*)initSnippet:(id)textView
// イニシャライザ
//------------------------------------------------------
{
    CESnippet *ret = [[CESnippet alloc] init];
    
    [ret initInstance:textView];
    
    return ret;
}

//------------------------------------------------------
- (void)initInstance:(id)textView
// インスタンスの初期化処理を行う
//------------------------------------------------------
{
	LOG(@"initInstance:");
    
    //------------------------------------------------------
	// 初期化処理
    //------------------------------------------------------

	ceTextView = textView;
    
    theTabTrigger = [[NSMutableString string] retain];
    
    // プレイスホルダ移動モードを NO に
    _isModeActivate = NO;
    
    // 初期設定ファイルがあるかどうかを確認する
    // 初期設定ファイルが無ければバンドルからコピーする
    
    // Snippets フォルダがあるかどうかを確認する
    NSFileManager *theFileManager = [NSFileManager defaultManager];
    NSString *theDirPath = [self pathOfSnippetsDirectory];
    BOOL theBoolIsDir = NO, theBoolCopyed = NO;
    BOOL theExists = [theFileManager fileExistsAtPath:theDirPath isDirectory:&theBoolIsDir];
    
    // フォルダが無ければバンドルからコピーする
    if (!theExists) {
        NSString *bundlePath = [[NSBundle bundleForClass:[CESnippet class]] resourcePath];
        NSString *snippetFolderPath = [NSString stringWithFormat:@"%@%@", bundlePath, @"/Snippets/"];
        NSError *error;
        theBoolCopyed = [theFileManager copyItemAtPath:snippetFolderPath
                                                toPath:theDirPath 
                                                 error:&error];
    }
    
    if ((!theExists) && (!theBoolCopyed)) {
        NSLog(@"*Error. ScriptMenu directory could not found.");
        return;
    }
    
    // スニペットを格納する辞書を初期化する
    snippetsDict = [NSMutableDictionary dictionary];
    
    // シンタックス別のスニペットを初期化する。ただしこの時点では失敗する。現在のシンタックスが決まっていないため。
    [self initSyntaxSnippet];
    
    LOG(@"initInstance end.");
}

//------------------------------------------------------
- (void)initSyntaxSnippet
// 選択されたシンタックスでスニペットを初期化する
//------------------------------------------------------
{
	// 現在のシンタックスを得る
    NSString* currentSyntax = [[[ceTextView delegate] performSelector:@selector(editorView)] performSelector:@selector(syntaxStyleNameToColoring)];
    LOG(@"--current: %@ --", currentSyntax);
    
    if (!currentSyntax) {
        
        // シンタックスの取得に失敗していれば処理を終了する
        LOG(@"syntax nothing");
        return;
    }
    
    if (!ceSyntax) {
    
        // シンタックスが未設定なら現在のシンタックスを得る
        ceSyntax = currentSyntax;
        LOG(@"init syntax");
    } else {
        
        // シンタックスが設定済みなら現在のシンタックスと変わっているかどうかを得る
        if ([ceSyntax isEqualToString:currentSyntax]) {

            // 変わってなければ初期化処理を無視する
            return;
        } else {
            
            // シンタックスを変更する
            ceSyntax = currentSyntax;
            LOG(@"change syntax");
        }
    }
    
    // 現在のシンタックスでフォルダを検索し、ファイルの内容で配列を初期化する
    NSFileManager *theFileManager = [NSFileManager defaultManager];
    NSString *theDirPath = [self pathOfSnippetsDirectory];
    BOOL theBoolIsDir = NO;
    
    // Snippets フォルダがあるかどうかを確認する
    if (![theFileManager fileExistsAtPath:theDirPath isDirectory:&theBoolIsDir]) {
        return;
    }
    if (!theBoolIsDir) return;
    
    // snippetsDict をクリアする
    [snippetsDict removeAllObjects];
    
    // シンタックス非依存の Snippets フォルダがあるかどうかを確認する
    NSString *theNoSyntaxDir = [NSString stringWithFormat:@"%@/%@", theDirPath, CES_NOSYNTAX_FOLDER];
    if ([theFileManager fileExistsAtPath:theNoSyntaxDir isDirectory:&theBoolIsDir]) {
    	if (theBoolIsDir) {
        	LOG(@"noSyntax Folder exist");
            // フォルダの内容を取り込む
            [self readSnippetFolder: theNoSyntaxDir];
            
            // スニペット更新用のコマンドを登録する
            [snippetsDict setObject:@"" forKey:CES_UPDATE_SNIPPETS_COMMAND];
        } else {
            LOG(@"noSyntax Folder no exist");
        }
    }
    
    // シンタックス対応の Snippets フォルダがあるかどうかを確認する
    NSString *theSyntaxDir = [NSString stringWithFormat:@"%@/%@", theDirPath, ceSyntax];
    if (ceSyntax) {
        if ([theFileManager fileExistsAtPath:theSyntaxDir isDirectory:&theBoolIsDir]) {
            if (theBoolIsDir) {
                LOG(@"Syntax Folder exist");
                // フォルダの内容を取り込む
                [self readSnippetFolder: theSyntaxDir];
            }
        } else {
            LOG(@"Syntax Folder no exist");
        }
    }
    
    // 完成した辞書を確保しておく
    [snippetsDict retain];
    
    #ifdef DEBUG
    // 辞書のキー（タブトリガー）をダンプする
    NSEnumerator* en = [snippetsDict keyEnumerator];
    id key;
    LOG(@"{");
    while (key = [en nextObject]) {
        LOG(@"    %@", [key description]);
    }
    LOG(@"}");
    #endif
}

//------------------------------------------------------
- (void)readSnippetFolder:(NSString*)theDirPath
// 指定されたフォルダ内のファイルでスニペットを初期化する
//------------------------------------------------------
{
	NSFileManager *theFileManager = [NSFileManager defaultManager];
	NSArray *theFiles;
    theFiles = [[NSFileManager defaultManager] directoryContentsAtPath:theDirPath];
    
    // フォルダの内容を全て辞書に格納するisReadableFileAtPath:
    NSString* 			snippetName = @"";
    NSMutableString* 	snippetText = @"";
    NSMutableString* 	theFilePath = @"";
    NSError*			err = nil;
    
    for (int i = 0; i < [theFiles count]; i++) {
    
    	theFilePath = [NSString stringWithFormat:@"%@/%@", theDirPath, [theFiles objectAtIndex:i]];
        
        // ファイル名を取得する。そのままタブトリガーになる
        snippetName = [theFilePath lastPathComponent];
        if ([snippetName isEqualToString:@".DS_Store"]) {
        	// .DS_Store は無視する
            continue;
        }
        
        snippetText = @"";
        
        // ファイルが読み込み可能かどうかを得る
        if ([theFileManager isReadableFileAtPath:theFilePath]) {
            
            LOG(theFilePath);
            // ファイル内容を得る
            snippetText = [NSMutableString stringWithContentsOfFile:theFilePath
                                                           encoding:NSUTF8StringEncoding
                                                              error:&err];
        }
        
        // 辞書に追加する
        if (snippetText) {
        	LOG(@"%@:%@", snippetName, snippetText);
            [snippetsDict setObject:snippetText forKey:snippetName];
        } else {
            LOG(@"%@:ファイル内容を取得できませんでした。:%@", snippetName, [err localizedDescription]);
        }
    }
}

//------------------------------------------------------
- (NSString *)pathOfSnippetsDirectory
// Snippetsファイル保存用ディレクトリを返す
//------------------------------------------------------
{
    NSString *outPath = [NSHomeDirectory( ) 
                         stringByAppendingPathComponent:CES_SNIPPETS_FOLDER];
    
    return outPath;
}

#pragma mark -- Variables --

//------------------------------------------------------
// スニペット中の変数を置き換える。
- (NSString *)translateVariables:(NSString*)snippet
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];
    
	// $XXX または ${XXX} 形式の文字列を抽出する正規表現
    NSString *regrex2 = @"(?<!\\\\)\\${1}[a-zA-Z_]+|(?<!\\\\)\\${1}\\{.+?\\}";
    OGRegularExpression* oniRegrex2 =
    [OGRegularExpression regularExpressionWithString:regrex2];
    
    // Snippet を正規表現で検索し、変数部分を得る
    NSEnumerator *theEnumerator;
    theEnumerator = [oniRegrex2 matchEnumeratorInString:snippet];
    NSArray *theMatchedAry = [oniRegrex2 allMatchesInString:snippet];
    
    if ([theMatchedAry count] == 0) {
    	LOG(@"Variables no exists.");
        return snippet;
    }
    
    // 全ての変数を置き換える
    NSMutableString* lastPostmatchString = [NSMutableString stringWithString:@""];
    for (OGRegularExpressionMatch* theMatch in theMatchedAry) {
        
    	LOG(@"Start replace -> %@", [theMatch matchedString]);
        
        // マッチ部分とマッチ部分の間の文字列を接続する
        [retString appendString:[theMatch stringBetweenMatchAndLastMatch]];
        
        // マッチ部分を変数として値を置き換え、接続する
        [retString appendString:[self replaceVariable:[theMatch matchedString]]];
        
        // マッチ部分の後ろの文字列を保存する
        [lastPostmatchString setString:[theMatch postmatchString]];
        
    }
    
    // 最後のマッチ部分の後ろの文字列を接続する
    [retString appendString:lastPostmatchString];
    
    return retString;
}

// -----------------------------------------------------
- (NSString *)replaceVariable:(NSString*)variableString
// 変数を置き換える
// -----------------------------------------------------
{
    NSArray *varSplitedBySlash;
    
    // 変数文字列から変数名だけを得る（すなわち最初の「$」または「${」に続く英数字を取得する）
    if ([variableString length] < 3) {
        return @"";
    }
    
	// $XXX または ${XXX:YYY} 形式の文字列から XXX だけを抽出する
    NSString *regrex = @"(?<!\\$)[a-zA-Z_]+|(?<!\\$\\{)[a-zA-Z_]+";
    OGRegularExpression* oniRegrex =
    	[OGRegularExpression regularExpressionWithString:regrex];
    
    OGRegularExpressionMatch* theMatch = [oniRegrex matchInString:variableString];
    NSString *varName = [theMatch matchedString];

    // ${XXX:YYY} 形式の文字列から XXX:YYY だけを抽出する
    OGRegularExpression* oniRegrex2 =
	    [OGRegularExpression regularExpressionWithString:@"[^\\$\\{]+.*\\b"];
    
    OGRegularExpressionMatch* theMatch2 = [oniRegrex2 matchInString:variableString];
    NSString *varFullName = [theMatch2 matchedString];

    // 処理タイプを得る
    // 処理タイプは以下の通り
    // 		variable
    // 		variable_with_default_value
    // 			${«variable»:«default value»}
    //		variable_with_regexp
    // 			${«variable»/«regexp»/«format»/«options»}
    //		TODO:★variable_with_file
    //			${PickRelativePath,title,filter}
    NSMutableString *procType = @"variable";
    
    // コロンで分割できるかどうかを確認する
    OGRegularExpression* oniRegrex3 =
	    [OGRegularExpression regularExpressionWithString:@"(?<!\\\\):"];
    NSArray *varSplitedByColon = [oniRegrex3 splitString:varFullName];
    if (1 < [varSplitedByColon count]) {
        
        procType = @"variable_with_default_value";
    } else {
    
        // スラッシュで分割できるかどうかを確認する
        OGRegularExpression* oniRegrex4 =
			[OGRegularExpression regularExpressionWithString:@"(?<!\\\\)/"];
        varSplitedBySlash = [oniRegrex4 splitString:varFullName];
        if (3 < [varSplitedBySlash count]) {
            procType = @"variable_with_regexp";
        }
    }
    
    // 変数を置き換える
    NSString *varReplacedString = [self varExecute:varName];
    
    // シンプルな置き換え
    if ([procType isEqualToString:@"variable"]) {
        return varReplacedString;
    }
    
    // 変数の置き換え結果が空文字列だった場合、且つ初期値がある場合、初期値を返す
    if ([procType isEqualToString:@"variable_with_default_value"]) {
    	if ([varReplacedString isEqualToString:@""]) {
        	// 末尾の } を削除する
            
        	NSString* defaultValue = [varSplitedByColon objectAtIndex:1];
            NSRange lastRange = NSMakeRange([defaultValue length] - 1, 1);
            NSString *lastChar = [NSString stringWithString:[defaultValue substringWithRange:lastRange]];
            if ([lastChar isEqualToString:@"}"]) {
                return [defaultValue substringWithRange:NSMakeRange(0, [defaultValue length] - 1)];
            } else {
                return defaultValue;
            }
            
        } else {
            return varReplacedString;
        }
    }
    
    // 正規表現だった場合、変数の置き換え結果に対して正規表現を適用する
    if ([procType isEqualToString:@"variable_with_regexp"]) {
        
        @try {
            // スラッシュで区切られた 2 番目の部分を正規表現として登録する
            OGRegularExpression* oniRegrex5 =
	        [OGRegularExpression regularExpressionWithString:[varSplitedBySlash objectAtIndex:1]];
            
            // 3 番目を置き換えの正規表現として置き換えを実行する
            return [oniRegrex5 replaceAllMatchesInString:varReplacedString 
                                              withString:[varSplitedBySlash objectAtIndex:2]
                                                 options:0];
        }
        @catch (NSException * e) {
        	return varReplacedString;
        }
    }
    
    // どれでもない場合
    return @"";
}

// -----------------------------------------------------
- (NSString *)varExecute:(NSString*)varName
// 変数を、実際の値に置き換える
// -----------------------------------------------------
{
	id ceEditorView = [[ceTextView delegate] performSelector:@selector(editorView)];
    id ceDocument = [ceEditorView performSelector:@selector(document)];
    
    // ファイルパス情報
    NSURL * fileUrl = [ceDocument performSelector:@selector(fileURL)];
    
    if ([varName isEqualToString:@"Path"] || [varName isEqualToString:@"TM_FILEPATH"]) {
    	// 現在のファイルの完全パス名。
        if (nil == fileUrl) return @"";
    	return [fileUrl path];
    } else if ([varName isEqualToString:@"Dir"] || [varName isEqualToString:@"TM_DIRECTORY"]) {
    	// 現在のファイルのディレクトリ名。
        if (nil == fileUrl) return @"";
        return [[fileUrl path] stringByDeletingLastPathComponent];
    } else if ([varName isEqualToString:@"Filename"]) {
    	// ファイル拡張子を除くファイル名。
        if (nil == fileUrl) return @"";
        return [[[fileUrl path] lastPathComponent] stringByDeletingPathExtension];
    } else if ([varName isEqualToString:@"FilenameEx"] || [varName isEqualToString:@"TM_FILENAME"]) {
    	// ファイル拡張子を含むファイル名。
        if (nil == fileUrl) return @"";
        return [[fileUrl path] lastPathComponent];
    } else if ([varName isEqualToString:@"Ext"]) {
    	// ファイル拡張子。
        if (nil == fileUrl) return @"";
    	return [[fileUrl path] pathExtension];
    }
    
    // 日時情報
    NSDate *now = [[NSDate date] retain];
    
    if ([varName isEqualToString:@"Date"]) {
    	// 今日の日付。
        return [now descriptionWithCalendarFormat:@"%Y/%m/%d" timeZone:nil locale:[NSLocale systemLocale]];
    } else if ([varName isEqualToString:@"Time"]) {
    	// 現在の時刻。
        return [now descriptionWithCalendarFormat:@"%H:%M:%S" timeZone:nil locale:[NSLocale systemLocale]];
    }
	
    // ドキュメント情報
    if ([varName isEqualToString:@"CurCol"] || [varName isEqualToString:@"TM_LINE_INDEX"]) {
    	// カーソル位置の論理桁番号。
        return [NSString stringWithFormat:@"%d", [ceEditorView currentColumn] - [theTabTrigger length]];
    } else if ([varName isEqualToString:@"CurLine"] || [varName isEqualToString:@"TM_LINE_NUMBER"]) {
    	// カーソル位置の論理桁番号。
        return [NSString stringWithFormat:@"%d", [ceEditorView currentLineNum]];
    } else if ([varName isEqualToString:@"SelText"] || [varName isEqualToString:@"TM_SELECTED_TEXT"]) {
    	// 選択テキスト
        return [[ceTextView string] substringWithRange:[ceTextView selectedRange]];
    } else if ([varName isEqualToString:@"LineText"] || [varName isEqualToString:@"TM_CURRENT_LINE"]) {
    	// 現在行のテキスト（省略形抜き）
        NSMutableString* lineText = [NSMutableString stringWithString:[ceEditorView currentLineText]];
        [lineText replaceCharactersInRange:NSMakeRange([ceEditorView currentColumn] - [theTabTrigger length],[theTabTrigger length]) withString:@""];
		return lineText;
    } else if ([varName isEqualToString:@"WordText"] || [varName isEqualToString:@"TM_CURRENT_WORD"]) {
        // 単語を分割する
        NSMutableString* lineText = [NSMutableString stringWithString:[ceEditorView currentLineText]];
        [lineText replaceCharactersInRange:NSMakeRange([ceEditorView currentColumn] - [theTabTrigger length],[theTabTrigger length]) withString:@""];
        
        OGRegularExpression* oniRegrex = [OGRegularExpression regularExpressionWithString:@"\\b[^\\s.,;:・。、]+\\b"];
        NSArray *theMatchedAry = [oniRegrex allMatchesInString:lineText];
        NSUInteger theCursorLoc = [ceEditorView currentColumn];
        
        for (OGRegularExpressionMatch* theWord in theMatchedAry) {
            
            // 単語の位置（範囲）がカーソル位置を含んでいるかどうかをチェックする
            NSRange theWordRange = [theWord rangeOfMatchedString];
            if (theWordRange.location <= theCursorLoc && theCursorLoc <= theWordRange.location + theWordRange.length) {
                return [theWord matchedString];
            }
        }

        return @"";
    }
        
    // HTML meta タグで使用される Charset。
    if ([varName isEqualToString:@"Charset"]) {
        
        id enc = [ceDocument performSelector:@selector(encodingCode)];
        
        return [NSString stringWithFormat:@"%@", CFStringConvertEncodingToIANACharSetName(CFStringConvertNSStringEncodingToEncoding((NSStringEncoding)enc))];
    }


	// CotEditor 設定にもとづく情報
    id theConfigValues = [[NSUserDefaultsController sharedUserDefaultsController] values];
    
    if ([varName isEqualToString:@"TabSize"]  || [varName isEqualToString:@"TM_TAB_SIZE"] ||
    	[varName isEqualToString:@"TabWidth"] || [varName isEqualToString:@"IndentSize"]) {
        
    	// タブのサイズ。CotEditor の場合、インデントのサイズでもある
        return [[theConfigValues valueForKey:k_key_tabWidth] description];
    } else if ([varName isEqualToString:@"UseSpacesForTabs"] || [varName isEqualToString:@"TM_SOFT_TABS"]) {
        
        // [タブの代わりに空白を挿入] が有効なら 1、そうでなければ 0。
        return [[theConfigValues valueForKey:k_key_autoExpandTab] description];
    } else if ([varName isEqualToString:@"AutoIndent"]) {
    	
        // 自動インデントが有効なら 1、そうでなければ 0。
        return [[theConfigValues valueForKey:k_key_autoIndent] description];
    } else if ([varName isEqualToString:@"AppVersion"]) {
        
        // CotEditor のバージョン
        return [[NSBundle mainBundle] objectForInfoDictionaryKey: @"CFBundleVersion"];
    } else if ([varName isEqualToString:@"PluginVersion"]) {
        
        // Snippet SIMBL プラグインのバージョン
        return [[NSBundle bundleForClass:[self class]] objectForInfoDictionaryKey: @"CFBundleVersion"];
    } else if ([varName isEqualToString:@"Indent"]) {
    	LOG_INT([theConfigValues valueForKey:k_key_autoExpandTab]);
        // 現在のインデント。タブ、または[タブの代わりに空白を挿入] が有効なら空白。
        if ([[[theConfigValues valueForKey:k_key_autoExpandTab] description] isEqualToString:@"1"]) {
        	
            // タブの幅分、空白を連結する
            NSMutableString* spc = [NSMutableString stringWithString:@""];
            int tabW = [[[theConfigValues valueForKey:k_key_tabWidth] description] integerValue];
            for (int i = 0; i < tabW; i++) {
                
                [spc appendString:@" "];
            }
            
            return [spc description];
        } else {
            return @"\t";
        }
    } else if ([varName isEqualToString:@"CurrentLineIndent"]) {
    
    	LOG(@"test");
    	// 現在行のインデント（行頭からの空白文字全て）
    	// 現在行のテキスト（省略形抜き）から先頭の空白文字を取得する
        NSMutableString* lineText = [NSMutableString stringWithString:[ceEditorView currentLineText]];
        [lineText replaceCharactersInRange:NSMakeRange([ceEditorView currentColumn] - [theTabTrigger length],[theTabTrigger length]) withString:@""];
        
        OGRegularExpression* oniRegrex = [OGRegularExpression regularExpressionWithString:@"^\\s+"];
        OGRegularExpressionMatch* theMatche = [oniRegrex matchInString:lineText];
        if (nil == theMatche) {
            return @"";
        }
        
        return [theMatche matchedString];
    }
    
    return @"";
}

@end




























