//
//  CESnippet.m
//  CESnippet
//
//  Created by 舘 康一 on 11/09/05.
//  Copyright 2011 Odoru-saboten Workshop. All rights reserved.
//

#import "CESnippet.h"
#import <OgreKit/OgreKit.h>

@interface CESnippet (Private)
- (NSString *)pathOfSnippetsDirectory;
- (void)initInstance:(id)textView;
- (void)initSyntaxSnippet;
- (void)readSnippetFolder:(NSString*)theDirPath;
- (NSString *)findAbbreviation;
- (BOOL)expandAbbreviation:(NSString*)abbreviation;
- (NSString *)translateShellScript:(NSString*)snippet;
- (NSString *)executeShellScript:(NSString *)commandBaseString;
- (NSString *)translateVariables:(NSString*)snippet;
- (NSString *)replaceVariable:(NSString*)variableString;
- (NSString *)translateTabStop:(NSString*)snippet;
- (NSString *)varExecute:(NSString*)varName;
@end

@implementation CESnippet

//------------------------------------------------------
+ (CESnippet*)initSnippet:(id)textView
// イニシャライザ
//------------------------------------------------------
{
    CESnippet *ret = [[CESnippet alloc] init];
    
    [ret initInstance:textView];
    
    return ret;
}

#pragma mark Public Method

//------------------------------------------------------
- (BOOL)modeActivate
// 現在 Snippet モードになっているかどうかを返す
//------------------------------------------------------
{
    return _isModeActivate;
}

//------------------------------------------------------
- (BOOL)executeSnippet:(id)sender
// スニペットを実行する
//------------------------------------------------------
{
    LOG(@"executeSnippet!");
    
    //-----------------------------------------------
    // Snippet モードになっていれば「次のプレイスホルダー」を実行する
    // Shiftキーが押されていたかどうかで挙動が違うが、その判定は呼び出し先のメソッドで行う。
    //-----------------------------------------------
    if (_isModeActivate) {
        // [self nextPlaceHolder:sender];
        
        // プレイスホルダー遷移処理を実行したため「実行した」を返し、タブ入力をキャンセルさせる
        return CES_SNIPPET_EXECUTED;
    }
	// Snippet モードでなければ以降の処理を行い、省略形の展開を試みる
    //-----------------------------------------------
    
    //-----------------------------------------------
    // 現在のシンタックスが取得できていなければスニペットの初期化が終わっていないので先に初期化をする。
    //-----------------------------------------------
    if (!ceSyntax) {
        [self initSyntaxSnippet];
    }
    
    //-----------------------------------------------
    // 現在のキャレット位置から前方検索し、省略形を探す。
    //-----------------------------------------------
    NSString* abbreviation = [self findAbbreviation];
    LOG(abbreviation);
    if ([abbreviation isEqualToString:@""]) {
    	// 省略形が見つからない場合「実行していない」を返す。これで通常のタブ入力となる。
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    //-----------------------------------------------
    // 省略形が、スニペットの初期化を促すものであれば初期化を行う
    //-----------------------------------------------
    if ([abbreviation isEqualToString:CES_UPDATE_SNIPPETS_COMMAND]) {
    	LOG(@"UPDATE_SNIPPETS_COMMAND");
        [self initSyntaxSnippet];
    }
    
    //-----------------------------------------------
    // 省略形を展開する
    //-----------------------------------------------
    BOOL successAbbreviation = [self expandAbbreviation:abbreviation];
    if (!successAbbreviation) {
    	// 省略形の展開が失敗した場合「実行していない」を返す。これで通常のタブ入力となる。
        LOG(@"Failed expand.");
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    //-----------------------------------------------
    // いくつかのプレイスホルダーが見つかっていればモードを Snippet モードにする
    //-----------------------------------------------
    // havePlaceHolder
    // [self 
    // _isModeActivate = YES;
    
    
    //-----------------------------------------------
    // 省略形が見つからない等、スニペットの実行が無ければ「実行していない」を返す。
    // 処理がここまで進んでいればスニペットの展開に問題がないと判断し、タブ入力イベントを
    // キャンセルさせるため「実行した」を返す。
    //-----------------------------------------------
    return CES_SNIPPET_EXECUTED;
}

//------------------------------------------------------
- (void)dealloc
// インスタンスの終了処理を行う
//------------------------------------------------------
{
    [snippetsDict release];
    [super dealloc];
}

#pragma mark Private Method

//------------------------------------------------------
// 省略形を展開する
- (BOOL)expandAbbreviation:(NSString*)abbreviation
//------------------------------------------------------
{
    NSString* snippet = [snippetsDict objectForKey:abbreviation];
    
    // 省略形（トリガー）が存在するかどうかを調べる
    if (!snippet) {
    	LOG(@"Tabtrigger Not Found in Array.");
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    // ||||||||||||||||||||||||||||||||||||||||||
    // 優先順にスニペットを実行していく
    // ||||||||||||||||||||||||||||||||||||||||||
    resultString = [NSMutableString stringWithString:snippet];
    
    // ------------------------------------------
    // シェルの実行
    // ------------------------------------------
	[resultString setString:[self translateShellScript:resultString]];
    if (nil == resultString) {
    	LOG_TRACE;
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    // ------------------------------------------
    // 変数の置き換え
    // ------------------------------------------
    [resultString setString:[self translateVariables:resultString]];
    if (nil == resultString) {
    	LOG_TRACE;
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    // ------------------------------------------
    // プレースホルダの置き換え
    // 置き換え後に各プレースホルダの位置を保存する
    // ------------------------------------------
    
    // タブストップ（$0のみ）を置き換え
	[resultString setString:[self translateTabStop:resultString]];
    if (nil == resultString) {
    	LOG_TRACE;
        return CES_SNIPPET_NOT_EXECUTED;
    }
    
    
    // ||||||||||||||||||||||||||||||||||||||||||
    // スニペットが完成したら省略形を置き換える
    // ||||||||||||||||||||||||||||||||||||||||||
    [ceTextView setSelectedRange:abbreviationRange];
    [ceTextView insertText:resultString];
    
    // タブストップが指定されていればその位置を選択状態とする
    if ((int)0 <= (int)cursorPositionRange.location) {
        [ceTextView setSelectedRange:cursorPositionRange];
    }
    
    // ここまで来ていれば省略形の展開を終えているはずなので「実行した」ことを呼び出し元に知らせる
    return CES_SNIPPET_EXECUTED;
}

//------------------------------------------------------
// スニペット中のタブストップ（$0）を置き換え、その位置を
// cursorPositionRange に保存する。
// ※ ${0:stop_pos} にはまだ対応していない。
- (NSString *)translateTabStop:(NSString*)snippet
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];
    
	// $0 に囲まれた文字列を抽出する正規表現
    NSString *regrex2 = @"(?<!\\\\)(\\$0){1}\\b";
    OGRegularExpression* oniRegrex2 =
    [OGRegularExpression regularExpressionWithString:regrex2];
    
    // Snippet を正規表現で検索し、最初のタブストップの位置を得る
    OGRegularExpressionMatch* theMatch = [oniRegrex2 matchInString:snippet];
    if (theMatch == nil) {
    	// マッチしなければ何もしない
        cursorPositionRange.location = -1;
        return snippet;
    }
    
    // $0 の位置を保存
    cursorPositionRange = [theMatch rangeOfMatchedString];
    // 置き換えるので長さはゼロに
    cursorPositionRange.length = 0;
    // 置き換え位置を、テキスト全体の位置に
    cursorPositionRange.location += abbreviationRange.location;
    
    // $0 を置き換え
    [retString appendString:[oniRegrex2 replaceFirstMatchInString:snippet
                                                       withString:@""]];
    return retString;
}

//------------------------------------------------------
// スニペット中の変数を置き換える。
- (NSString *)translateVariables:(NSString*)snippet
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];
    
	// $XXX または ${XXX} 形式の文字列を抽出する正規表現
    NSString *regrex2 = @"(?<!\\\\)\\${1}[a-zA-Z_]+|(?<!\\\\)\\${1}\\{.+?\\}";
    OGRegularExpression* oniRegrex2 =
    [OGRegularExpression regularExpressionWithString:regrex2];
    
    // Snippet を正規表現で検索し、変数部分を得る
    NSEnumerator *theEnumerator;
    theEnumerator = [oniRegrex2 matchEnumeratorInString:snippet];
    NSArray *theMatchedAry = [oniRegrex2 allMatchesInString:snippet];
    
    if ([theMatchedAry count] == 0) {
    	LOG(@"Variables no exists.");
        return snippet;
    }
    
    // 全ての変数を置き換える
    NSMutableString* lastPostmatchString = [NSMutableString stringWithString:@""];
    for (OGRegularExpressionMatch* theMatch in theMatchedAry) {
        
    	LOG(@"Start replace -> %@", [theMatch matchedString]);
        
        // マッチ部分とマッチ部分の間の文字列を接続する
        [retString appendString:[theMatch stringBetweenMatchAndLastMatch]];
        
        // マッチ部分を変数として値を置き換え、接続する
        [retString appendString:[self replaceVariable:[theMatch matchedString]]];
        
        // マッチ部分の後ろの文字列を保存する
        [lastPostmatchString setString:[theMatch postmatchString]];
    }
    // 最後のマッチ部分の後ろの文字列を接続する
    [retString appendString:lastPostmatchString];
    
    return retString;
}

//------------------------------------------------------
// スニペットの中のシェルスクリプトを実行する
// スクリプト部分を実行結果で置き換えるが、実行に失敗した場合
// 空文字列で置き換える。
- (NSString *)translateShellScript:(NSString*)snippet
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];

	// ` ` に囲まれた文字列を抽出する正規表現
    NSString *regrex2 = @"(?<!\\\\)`{1}[^`]*(?<!\\\\)`{1}";
    OGRegularExpression* oniRegrex2 =
    	[OGRegularExpression regularExpressionWithString:regrex2];
    
    // Snippet を正規表現で検索し、シェルスクリプト部分を得る
    NSEnumerator *theEnumerator;
    theEnumerator = [oniRegrex2 matchEnumeratorInString:snippet];
    NSArray *theMatchedAry = [oniRegrex2 allMatchesInString:snippet];
    
    if ([theMatchedAry count] == 0) {
    	LOG(@"Shell no execute.");
        return snippet;
    }
    
    // 全てのシェルを置き換える
    NSMutableString* lastPostmatchString = [NSMutableString stringWithString:@""];
    for (OGRegularExpressionMatch* theMatch in theMatchedAry) {
    
    	LOG(@"Start execute -> %@", [theMatch matchedString]);
    
        // マッチ部分とマッチ部分の間の文字列を接続する
        [retString appendString:[theMatch stringBetweenMatchAndLastMatch]];
    
        // マッチ部分をシェルスクリプトとして実行し、接続する
        [retString appendString:[self executeShellScript:[theMatch matchedString]]];
        
        // マッチ部分の後ろの文字列を保存する
        [lastPostmatchString setString:[theMatch postmatchString]];
    }
    // 最後のマッチ部分の後ろの文字列を接続する
    [retString appendString:lastPostmatchString];
    
    return retString;
}

//------------------------------------------------------
// シェルスクリプトを実行する
- (NSString *)executeShellScript:(NSString *)commandBaseString
//------------------------------------------------------
{
    // 返す文字列
    NSMutableString* retString = [NSMutableString stringWithString:@""];
    
    // コマンド部分を取り出す（すなわち最初と最後の文字「`」を削除する）
    if ([commandBaseString length] < 3) {
        return @"";
    }
    NSRange commandRange = NSMakeRange(1, [commandBaseString length] - 2);
    NSString *commandString = [NSString stringWithString:[commandBaseString substringWithRange:commandRange]];
    
    // コマンド部分を実行する
    NSTask *task  = [[NSTask alloc] init];
    NSPipe *pipe  = [[NSPipe alloc] init];
    
    [task setLaunchPath: @"/bin/sh"];
    [task setArguments: [NSArray arrayWithObjects: @"-c", commandString, nil]];
    
    @try {
    	LOG(commandString);
        
        [task setStandardOutput: pipe];
        [task launch];
        
        LOG(@"task complete");
        
        NSFileHandle *handle = [pipe fileHandleForReading];
        NSData *data = [handle  readDataToEndOfFile];
        
        LOG(@"data complete");
        
        [retString setString:[[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease]];
        
        // 最後の文字が改行コードの場合、それを削除する
        NSRange lastRange = NSMakeRange([retString length] - 1, 1);
        NSString *lastChar = [NSString stringWithString:[retString substringWithRange:lastRange]];
        if ([lastChar isEqualToString:@"\n"]) {
            [retString setString:
            	[retString substringWithRange:NSMakeRange(0, [retString length] -1)]];
        }
        
        LOG(retString);
    }
    @catch (NSException * e) {
    	// 実行に失敗したら空文字列を返す。
        NSLog([e description]);
        [retString setString: @""];
    }
    @finally {
        [task release];
        [pipe release];
    }
    return retString;
}

//------------------------------------------------------
// 「省略形」を探す
- (NSString *)findAbbreviation
//------------------------------------------------------
{
    // 現在のキャレット位置を得る
    NSRange theSelectedRange = [ceTextView selectedRange];
    LOG_RANGE(theSelectedRange);
    
    // 現在行の開始位置から現在のキャレットまでの文字列を得る
    NSRange theSelectedLineRange = [[ceTextView string] lineRangeForRange:theSelectedRange];
    LOG_RANGE(theSelectedLineRange);
    
    NSUInteger theLength = theSelectedRange.location - theSelectedLineRange.location;
    NSString* theSelectedString = [[ceTextView string] substringWithRange:NSMakeRange(theSelectedLineRange.location, theLength)];
    LOG(theSelectedString);
    
    // 得られた文字列を空白で分割し、最後に得られた文字列を「省略形」と看做す
    Class oni = objc_getClass("OGRegularExpression");
    @try {
        NSString *regrex = @"[\\s]+";
        id oniRegrex = 
	    	[oni performSelector:@selector(regularExpressionWithString:) withObject:regrex];
        NSArray* ary = 
    	    [oniRegrex performSelector:@selector(splitString:) withObject:theSelectedString];
        
        // 省略形が無ければ空文字列を返す。これで通常のタブ入力となる。
        if ([ary count] <= 0) {
            return @"";
        }
        
        // 省略形が見つかった位置(NSRange)を保存する
        abbreviationRange = NSMakeRange(theSelectedRange.location
        	- [[ary objectAtIndex:[ary count]-1] length], [[ary objectAtIndex:[ary count]-1] length]);
        LOG_RANGE(abbreviationRange);
        LOG_INT([[ary objectAtIndex:[ary count]-1] length]);
        
        // 得られた「省略形」を返す
        return [ary objectAtIndex:[ary count]-1];
    }
    @catch (NSException * e) {
    	// ogreKit が見つからない等、例外が発生したら空文字列を返す。これで
        // 通常のタブ入力となる。
        NSLog([e description]);
        return @"";
    }
    
    return @"";
}

#pragma mark -- init --

//------------------------------------------------------
- (void)initInstance:(id)textView
// インスタンスの初期化処理を行う
//------------------------------------------------------
{
	LOG(@"initInstance:");
    
    //------------------------------------------------------
	// 初期化処理
    //------------------------------------------------------

	ceTextView = textView;
    
    // スニペットモードを NO に
    _isModeActivate = NO;
    
    // 初期設定ファイルがあるかどうかを確認する
    // 初期設定ファイルが無ければバンドルからコピーする
    
    // Snippets フォルダがあるかどうかを確認する
    NSFileManager *theFileManager = [NSFileManager defaultManager];
    NSString *theDirPath = [self pathOfSnippetsDirectory];
    BOOL theBoolIsDir = NO, theBoolCopyed = NO;
    BOOL theExists = [theFileManager fileExistsAtPath:theDirPath isDirectory:&theBoolIsDir];
    
    // フォルダが無ければバンドルからコピーする
    if (!theExists) {
        NSString *bundlePath = [[NSBundle bundleForClass:[CESnippet class]] resourcePath];
        NSString *snippetFolderPath = [NSString stringWithFormat:@"%@%@", bundlePath, @"/Snippets/"];
        NSError *error;
        theBoolCopyed = [theFileManager copyItemAtPath:snippetFolderPath
                                                toPath:theDirPath 
                                                 error:&error];
    }
    
    if ((!theExists) && (!theBoolCopyed)) {
        NSLog(@"*Error. ScriptMenu directory could not found.");
        return;
    }
    
    // スニペットを格納する辞書を初期化する
    snippetsDict = [NSMutableDictionary dictionary];
    
    // シンタックス別のスニペットを初期化する。ただしこの時点では失敗する。現在のシンタックスが決まっていないため。
    [self initSyntaxSnippet];
    
    LOG(@"initInstance end.");
}

//------------------------------------------------------
- (void)initSyntaxSnippet
// 選択されたシンタックスでスニペットを初期化する
//------------------------------------------------------
{
    if (!ceSyntax) {
        // 現在のシンタックスを得る
        ceSyntax = [[[ceTextView delegate] performSelector:@selector(editorView)] performSelector:@selector(syntaxStyleNameToColoring)];
        
        if (!ceSyntax) {
        	
        	// シンタックスの取得に失敗していれば処理を終了する
            LOG(@"syntax nothing");
        	return;
        }
    }
    LOG(@"--%@--", ceSyntax);
    
    // 現在のシンタックスでフォルダを検索し、ファイルの内容で配列を初期化する
    NSFileManager *theFileManager = [NSFileManager defaultManager];
    NSString *theDirPath = [self pathOfSnippetsDirectory];
    BOOL theBoolIsDir = NO;
    
    // Snippets フォルダがあるかどうかを確認する
    if (![theFileManager fileExistsAtPath:theDirPath isDirectory:&theBoolIsDir]) {
        return;
    }
    if (!theBoolIsDir) return;
    
    // snippetsDict をクリアする
    [snippetsDict removeAllObjects];
    
    // シンタックス非依存の Snippets フォルダがあるかどうかを確認する
    NSString *theNoSyntaxDir = [NSString stringWithFormat:@"%@/%@", theDirPath, CES_NOSYNTAX_FOLDER];
    if ([theFileManager fileExistsAtPath:theNoSyntaxDir isDirectory:&theBoolIsDir]) {
    	if (theBoolIsDir) {
        	LOG(@"noSyntax Folder exist");
            // フォルダの内容を取り込む
            [self readSnippetFolder: theNoSyntaxDir];
            
            // スニペット更新用のコマンドを登録する
            [snippetsDict setObject:@"" forKey:CES_UPDATE_SNIPPETS_COMMAND];
        }
    }
    
    // シンタックス対応の Snippets フォルダがあるかどうかを確認する
    NSString *theSyntaxDir = [NSString stringWithFormat:@"%@/%@", theDirPath, ceSyntax];
    if (ceSyntax) {
        if ([theFileManager fileExistsAtPath:theSyntaxDir isDirectory:&theBoolIsDir]) {
            if (theBoolIsDir) {
                LOG(@"Syntax Folder exist");
                // フォルダの内容を取り込む
                [self readSnippetFolder: theSyntaxDir];
            }
        }
    }
    
    // 完成した辞書を確保しておく
    [snippetsDict retain];
    
    LOG([snippetsDict description]);
}

//------------------------------------------------------
- (void)readSnippetFolder:(NSString*)theDirPath
// 指定されたフォルダ内のファイルでスニペットを初期化する
//------------------------------------------------------
{
	NSFileManager *theFileManager = [NSFileManager defaultManager];
	NSArray *theFiles;
    theFiles = [[NSFileManager defaultManager] directoryContentsAtPath:theDirPath];
    
    // フォルダの内容を全て辞書に格納するisReadableFileAtPath:
    NSString* 			snippetName = @"";
    NSMutableString* 	snippetText = @"";
    NSMutableString* 	theFilePath = @"";
    
    for (int i = 0; i < [theFiles count]; i++) {
    
    	theFilePath = [NSString stringWithFormat:@"%@/%@", theDirPath, [theFiles objectAtIndex:i]];
        
        // ファイル名を取得する。そのままタブトリガーになる
        snippetName = [theFilePath lastPathComponent];
        if ([snippetName isEqualToString:@".DS_Store"]) {
            continue;
        }
        
        snippetText = @"";
        
        // ファイルが読み込み可能かどうかを得る
        if ([theFileManager isReadableFileAtPath:theFilePath]) {
            
            LOG(theFilePath);
            // ファイル内容を得る
            snippetText = [NSMutableString stringWithContentsOfFile:theFilePath];
        }
        
        // 辞書に追加する
        [snippetsDict setObject:snippetText forKey:snippetName];
    }
}

//------------------------------------------------------
- (NSString *)pathOfSnippetsDirectory
// Snippetsファイル保存用ディレクトリを返す
//------------------------------------------------------
{
    NSString *outPath = [NSHomeDirectory( ) 
                         stringByAppendingPathComponent:CES_SNIPPETS_FOLDER];
    
    return outPath;
}

#pragma mark -- Variables --

// -----------------------------------------------------
- (NSString *)replaceVariable:(NSString*)variableString
// 変数を置き換える
// -----------------------------------------------------
{
    // 変数文字列から変数名だけを得る（すなわち最初の「$」または「${」に続く英数字を取得する）
    if ([variableString length] < 3) {
        return @"";
    }
    
	// $XXX または ${XXX:YYY} 形式の文字列から XXX だけを抽出する
    NSString *regrex = @"(?<!\\$)[a-zA-Z_]+|(?<!\\$\\{)[a-zA-Z_]+";
    OGRegularExpression* oniRegrex =
    	[OGRegularExpression regularExpressionWithString:regrex];
    
    OGRegularExpressionMatch* theMatch = [oniRegrex matchInString:variableString];
    NSString *varName = [theMatch matchedString];
    
    // ${XXX:YYY} 形式の文字列から XXX:YYY だけを抽出する
    OGRegularExpression* oniRegrex2 =
	    [OGRegularExpression regularExpressionWithString:@"[^\\$\\{]+.*\\b"];
    
    OGRegularExpressionMatch* theMatch2 = [oniRegrex2 matchInString:variableString];
    NSString *varFullName = [theMatch2 matchedString];
    
    // 処理タイプを得る
    // 処理タイプは以下の通り
    // 		variable
    // 		variable_with_default_value
    // 			${«variable»:«default value»}
    //		variable_with_regexp
    // 			${«variable»/«regexp»/«format»/«options»}
    
    NSMutableString *procType = @"variable";
    
    // コロンで分割できるかどうかを確認する
    OGRegularExpression* oniRegrex3 =
	    [OGRegularExpression regularExpressionWithString:@"(?<!\\):"];
    NSArray *varSplitedByColon = [oniRegrex3 splitString:varFullName];
    if (0 < [varSplitedByColon count]) {
        [procType setString:@"variable_with_default_value"];
    } else {
        
        // スラッシュで分割できるかどうかを確認する
        OGRegularExpression* oniRegrex4 =
		[OGRegularExpression regularExpressionWithString:@"(?<!\\)/"];
        NSArray *varSplitedBySlash = [oniRegrex4 splitString:varFullName];
        if (0 < [varSplitedBySlash count]) {
            [procType setString:@"variable_with_regexp"];
        }
    }
    
    // 変数を置き換える
    NSString *varReplacedString = [self varExecute:varName];
    
    // シンプルな置き換え
    if ([procType isEqualToString:@"variable"]) {
        return varReplacedString;
    }
    
    // 変数の置き換え結果が空文字列だった場合、且つ初期値がある場合、初期値を返す
    if ([procType isEqualToString:@"variable_with_default_value"]) {
    	if ([varReplacedString isEqualToString:@""]) {
            return [varSplitedByColon objectAtIndex:1];
        } else {
            return varReplacedString;
        }
    }
    
    // 正規表現だった場合、変数の置き換え結果に対して正規表現を適用する
    if ([procType isEqualToString:@"variable_with_regexp"]) {
        
        // TODO:ちょっとめんどくさいので後日...
    }
    
    // どれでもない場合
    return @"";
}
@end































